// Basic navigation for sighted users, supports 8 directions and up/down the structure. Uses jQuery 2.14 functions.

//A map capable of using objects as keys. 
function dmap() {
	this.keys = [];
	this.values = [];
}

// Put something in the map.
dmap.prototype.put = function(key, value) {
	this.keys.push(key);
	this.values.push(value);
}

// Get something from the map.
dmap.prototype.get = function(key) {
	var index = this.keys.indexOf(key);
	return (index == -1) ? null : this.values[index];
}

// Returns true if the map contains the key.
dmap.prototype.contains = function(key) {
	return this.keys.indexOf(key) != -1;
}

// Basic Navigation for sighted users. Supports movement in 8 directions and up/down the structure.
// parentId is the id of the HTML element (often a <div> containing the SVG
function svgNavigation(parentId) {

	var id = '#' + parentId; // jQuery Id for the container
	this.$svg = $(id).find('svg'); // The SVG belonging to the parent container

	// numeric keys are covered with e, nw, ne, w, sw, se, s
	this.keys = {
		tab : 9,
		enter : 13,
		escape : 27,
		space : 32,
		pageup : 33,
		pagedown : 34,
		end : 35,
		home : 36,  
		left : 37,
		up : 38,
		right : 39,
		down : 40,
		asterisk : 106,
		e : 54,
		nw : 55,
		n : 56,
		ne : 57,
		w : 58,
		sw : 49,
		se : 51,
		s : 50,
		n1 : 97,
		n2: 98,
		n3: 99,
		n4: 100,
		n6: 102,
		n7: 103,
		n8: 104,
		n9: 105
	};
	
	this.svgNavUtil = new svgNavUtil();  // Utility functions
	// initialize the navigations
	this.init();
	// bind event handlers
	this.bindHandlers();
}

// Builds the navigation.
svgNavigation.prototype.init = function() {
	// Find navigable objects
	
	// Objects with roles should be navigable, unless role is none 
	var $roleNone = this.$svg.find("[role='none']");
	var $withRoles = this.$svg.find('[role]').not($roleNone);

	// Objects with a first child of <title> or <desc> should be navigable,
	// unless role of none 
	var $hasTitles = this.$svg.find(':first-child').filter('title').parent();
	var $hasDesc = this.$svg.find(':first-child').filter('desc').parent();

	// Objects with aria-(label, labeledby, describedby) should be navigable
	var $labeledBy = this.$svg.find('[aria-labeledby]');
	var $describedBy = this.$svg.find('[aria-describedby]');
	var $labels = this.$svg.find('[aria-label]');

	// everything navigable
	var $all = $withRoles.add($describedBy).add($labels).add($labeledBy).add($hasTitles).add($hasDesc).not($roleNone);

	// Groups with semantic meaning
	this.$groups = $all.filter('g');
	this.$groups = this.$groups.add(this.$svg.find('svg'));
	
	this.$navigableItems = $all;
	
	this.dmap = new dmap();  // Map of navigations, key is DOM nodes, value is navigation item.
	
	// Add atomic items to dmap, atomic items will be leaves in the structure (no children)
	var $atomicGroups = this.mapAtomicItems($all);
	
	// Build this navigation option
	this.buildNavigation($atomicGroups);
	
	// Selected is the navigation item for the DOM node with focus
	this.selected = this.getSvgNav();

} // end init()

// Creates an eightWayNavItem for atomic items, puts them in this.dmap
svgNavigation.prototype.mapAtomicItems = function($all) {

	// Find the atomic items
	// Find groups with atomic roles
	var $atomicGroups = this.$groups.filter("[role='symbol']");
	$atomicGroups = $atomicGroups.add(this.$groups.filter("[role='label']"));
	
	// Atomic items
	var $atomicItems = $all.not(this.$groups).not(this.$svg);
	var $atomics = $atomicItems.add($atomicGroups);
	
	var svg = this.$svg.get(0);	// DOM svg element

	// Create an eightWayNavItem for each atomic item, add to map
	var length = $atomics.length;
	for (var i = 0; i < length; i++) {
		var item = $atomics.get(i);
		var navItem = new eightWayNavItem(item, svg);
		this.dmap.put(item, navItem);
	}

	return $atomicGroups;  // Return groups with atomic roles
}

// Build the navigation
// Navigation structure consists of navFeatures and eightWayNavItems
// eightWayNavItem's represent a DOM element
// navFeatures retain point of regard for features and structures
svgNavigation.prototype.buildNavigation = function($atomicGroups) {
	var $used = $();  // jQuery object of used DOM elements
	var svg = this.$svg.get(0);  // SVG DOM element
	
	// We find the group and svg elements that have semantic meaning and will included in navigation
	// jQuery puts DOM nodes in DOM order in jQuery objects
	// We will go last to first through the groups building the navigation structure
	// the backward order ensures all the children for a semantic group are in the dmap when the 
	// navigation for the group is built
	var $semanticGroups = this.$groups.not($atomicGroups);
	var $semanticGroups = $semanticGroups.add(this.$svg);
	var max = $semanticGroups.length - 1;
	for (var i = max; i > -1; i--) {
		$descendants = this.makeSemanticGroupNavigation($semanticGroups.get(i), svg, $used);
		$used = $used.add($descendants);
	}

}

// Populates the navigation items for the group
// Initializes the siblings that can be navigated that belong to the group
svgNavigation.prototype.makeSemanticGroupNavigation = function(g, svg, $used) {
	var $g = $(g);
	var $descendants = $g.find(this.$navigableItems).not($used); // members for the group
	var role = this.svgNavUtil.calcRole(g);
	// calculate centers from descendants unless a graphics-figure role
	var centers = (role && role.indexOf('graphics-figure') == 0) ? null : this.centersFromDescendants($descendants);
	var centerCardinal = (centers) ? {
		x : centers.x,
		y : centers.y
	} : null;
	var centerDiagonal = (centers) ? {
		x : centers.dx,
		y : centers.dy
	} : null;
	
	var nav = new eightWayNavItem(g, svg, centerCardinal, centerDiagonal);
	this.dmap.put(g, nav);  // Add this group to the map, the group's parent will need it

	if ($descendants.length > 0) {
		this.addRelationships($descendants, nav);	
	}
	return $descendants; // Returns the DOM elements used in this group
}

// Add relationships that define the navigation behavior for the group.
svgNavigation.prototype.addRelationships = function($siblings, parentNav) {

	// Splitting siblings by role
	var roleMap = {};
	var length = $siblings.length;
	for (var i = 0; i < length; i++) {
		var domNode = $siblings.get(i);

		var role = this.svgNavUtil.calcRole(domNode);
		var array = roleMap[role];
		if (!array) {
			array = [];
			roleMap[role] = array;
		}
		array.push(domNode);
	}
	
	var count = Object.keys(roleMap).length;
	var childFeatures = [];

	// Navigation is for objects with the same role, if a group contains multiple roles,
	// then each role is considered a separate feature for navigation
	
	for ( var role in roleMap) {
		// Add sort arrays and sort siblings
		var array = roleMap[role];
		var ax = [];
		var ay = [];
		var dx = [];
		var dy = [];
		length = array.length;
		var nav = null;
		// Each role defines a separate feature for navigation 
		var feature = new navFeature(parentNav);
		childFeatures.push(feature);

		for(var i = 0; i < length; i++) {
			nav = this.dmap.get(array[i]);
			nav.setArrays(ax, ay, dx, dy, feature);
		}

		ax.sort(this.svgNavUtil.sortX);
		ay.sort(this.svgNavUtil.sortY);
		dx.sort(this.svgNavUtil.sortX);
		dy.sort(this.svgNavUtil.sortY);

		feature.pointOfRegard = ax[0].get8wayNavItem();
	}
	
	var max = childFeatures.length - 1;
	if (max > 0) {
		childFeatures[max].setNext(childFeatures[0]);
		for(var i = 0; i < max; i++) {
			childFeatures[i].setNext(childFeatures[i + 1]);
		}
	}
	
	if (parentNav) {
		parentNav.setChildWithRegard(childFeatures[0]); 
	}
}

svgNavigation.prototype.centersFromDescendants = function($descendants) {
	var length = $descendants.length;
	if (length == 0) {
		return null;
	}

	var totalcx = 0;
	var totalcy = 0;
	var totaldx = 0;
	var totaldy = 0;
	for (var i = 0; i < length; i++) {
		var navItem = this.dmap.get($descendants[i]);
		totalcx += navItem.cardinal.x;
		totalcy += navItem.cardinal.y;
		totaldx += navItem.diagonal.x;
		totaldy += navItem.diagonal.y;
	}

	// center for sorting
	var centers = {
		x : totalcx / length,
		y : totalcy / length,
		dx : totaldx / length,
		dy : totaldy / length
	};
	return centers;
}

//
//Function bindHandlers() is a member function to bind event handlers to the
//navigable items
//
//return N/A
//
svgNavigation.prototype.bindHandlers = function() {

	var thisObj = this;
	var $par = this.$svg.parent();
	$par.attr('tabindex', '0');

	// bind a click handler on each navigable item
	// this.$navigableItems.click(function(e) {
	// return thisObj.handleClick($(this), e);
	// });

	// bind a keydown handler to the SVG object
	$par.keydown(function(e) {
		return thisObj.handleKeyDown(e);
	});

	// bind a keypress handler to the SVG object
	// $par.keypress(function(e) {
	// return thisObj.handleKeyPress(e);
	// });

	// bind a focus handler to the SVG object

	$par.focus(function(e) {
		if (this.$activeItem) {
			this.$activeItem.attr('tabindex', '0');
			// This is a bad thing to do, but it shows navigation in Firefox and
			// Chrome
			this.$activeItem.attr('class', 'activeElem');
			return thisObj.handleFocus(e);
		}

	});
}

//
//Function handleKeyDown() is a member function to process keydown events for
//navigation
//@param (e object) e is the associated event object
//@return (boolean) returns false if consuming event; true if not
//
svgNavigation.prototype.handleKeyDown = function(e) {
	var caseFound = true;

	switch (e.keyCode) {
	case this.keys.space: {
		var selected = (this.selected) ? this.selected.nextFeature(!e.shiftKey) : this.getSvgNav();
		this.setActiveItem(selected);
		break;
	}
	case this.keys.left:
	case this.keys.w: 
	case this.keys.n4: {
		var selected = (this.selected) ? this.selected.west() : this.getSvgNav();
		this.setActiveItem(selected);
		break;
	}
	case this.keys.right:
	case this.keys.w: 
	case this.keys.n5: {
		var selected = (this.selected) ? this.selected.east() : this.getSvgNav();
		this.setActiveItem(selected);
		break;
	}
	case this.keys.up:
	case this.keys.n: 
	case this.keys.n8: {
		var selected = (this.selected) ? this.selected.north() : this.getSvgNav();
		this.setActiveItem(selected);
		break;
	}
	case this.keys.down:
	case this.keys.s: 
	case this.keys.n2: {
		var selected = (this.selected) ? this.selected.south() : this.getSvgNav();
		this.setActiveItem(selected);
		break;
	}

	case this.keys.nw:
	case this.keys.home: 
	case this.keys.n7: {
		var selected = (this.selected) ? this.selected.northwest() : this.getSvgNav();
		this.setActiveItem(selected);
		break;
	}
	case this.keys.se:
	case this.keys.pagedown: 
	case this.keys.n3: {
		var selected = (this.selected) ? this.selected.southeast() : this.getSvgNav();
		this.setActiveItem(selected);
		break;
	}
	case this.keys.sw:
	case this.keys.end:
	case this.keys.n1: {
		var selected = (this.selected) ? this.selected.southwest() : this.getSvgNav();
		this.setActiveItem(selected);
		break;
	}
	case this.keys.ne:
	case this.keys.pageup: 
	case this.keys.n9: {
		var selected = (this.selected) ? this.selected.northeast() : this.getSvgNav();
		this.setActiveItem(selected);
		break;
	}
	
	case this.keys.enter: {
		var selected = (this.selected) ? this.selected.down() : this.getSvgNav();
		this.setActiveItem(selected);
		break;
	}
	
	case this.keys.escape: {
		var selected = (this.selected) ? this.selected.up() : this.getSvgNav();
		this.setActiveItem(selected);
		break;
	}

	default:
		caseFound = false;
	}

	if (caseFound) {
		e.stopPropagation();
	}
	return !caseFound;

} // end handleKeyDown

// Set focus on the DOM elmement of the navItem
svgNavigation.prototype.setActiveItem = function(navItem) {
	this.selected = navItem;
	
	
	// update the active item and give it focus
	
	var item = (navItem) ? navItem.getItem() : null; // item is the DOM element
	this.$activeItem = $(item);			// jQuery object for the DOM element to get focus
	this.$navigableItems.attr('tabindex', '-1');
	// This is a bad thing to do, but it shows navigation in Firefox and Chrome
	this.$navigableItems.removeAttr("class");
	// removeClass is what we should be doing, but it does not affect the
	// outerHTML
	// which needs to happen for SVG to pick up the CSS changes
	// this.$navigableItems.removeClass("activeElem");
	this.$svg.attr('tabindex', '0');
	if (this.$activeItem) {
		this.$activeItem.attr('tabindex', '0');
		// This is a bad thing to do, but it shows navigation in Firefox and
		// Chrome
		this.$activeItem.attr('class', 'activeElem');
		// addClass is what we should be doing, but it does not affect the
		// outerHTML
		// which needs to happen for SVG to pick up the CSS changes
		// this.$activeItem.addClass("activeElem");
	}
}

// Get the SVG DOM element 
svgNavigation.prototype.getSvgNav = function() {
	return this.dmap.get(this.$svg.get(0));
}

//
//Function handleFocus() is a member function to process focus events.
//
//@param (e object) e is the associated event object
//
//@return (boolean) returns true
//
svgNavigation.prototype.handleFocus = function(e) {
	if (!this.selected) {
		this.selected = this.getSvgNav();
	}
	this.setActiveItem(this.selected);
	return true;

} // end handleFocus

//
//Function handleBlur() is a member function to process blur events.
//
//@param ($id object) $id is the jQuery id of the parent item firing event
//
//@param (e object) e is the associated event object
//
//@return (boolean) returns true
//
svgNavigation.prototype.handleBlur = function($id, e) {
	this.$navigableItems.attr('tabindex', '-1');
	this.$navigableItems.removeAttr("class");
	// removeClass is what we should be doing, but it does not affect the
	// outerHTML
	// which needs to happen for SVG to pick up the CSS changes
	// We should be using removeClass
	this.$navigableItems.removeClass('activeElem');
	this.$svg.attr('tabindex', '0');
	return true;

} // end handleBlur

